<!---------------------------------------------------------------------------->
<html>
<head>
<title> Sig++ example program: keyan.cpp </title>
<meta name="Document-Owner" content="Craig Sapp (craig@ccrma.stanford.edu)"> 
<meta name="Author" content="Craig Stuart Sapp (craig@ccrma.stanford.edu)">
<meta name="Creation-Date" content=" Tue May  9 19:38:11 PDT 2000 ">
<meta name="Revision-Date" content=" Sun May 18 15:48:39 PDT 2008
 ">
</head>
<!---------------------------------------------------------------------------->
<body bgcolor=#ffffff link=#223399 vlink=#223399 alink=#223399 hlink=#223399>    
<pre>
<font color=#229933>//</font>
<font color=#229933>// Programmer:    Craig Stuart Sapp &lt;craig@ccrma.stanford.edu&gt;</font>
<font color=#229933>// Creation Date: Sun May 16 12:54:28 PDT 1999</font>
<font color=#229933>// Last Modified: Sun May 16 16:06:30 PDT 1999</font>
<font color=#229933>// Filename:      .../improv/examples/synthImprov/keyan/keyan.cpp</font>
<font color=#229933>// Syntax:        C++; synthImprov 2.1</font>
<font color=#229933>//  </font>
<font color=#229933>// Description:   analyze the key of musical input and output the determined</font>
<font color=#229933>//                key as a low note</font>
<font color=#229933>//</font>
<font color=#229933>// References:    Key analysis method taken from the "key" program of Humdrum</font>
<font color=#229933>//                The Key analysis technique is described in the book:</font>
<font color=#229933>//                Krumhansel & Kessler</font>
<font color=#229933>//</font>


#include "<a href=http://sig.sapp.org/include/sig/synthImprov.h>synthImprov.h</a>"      
#include &lt;math.h&gt;

#ifndef OLDCPP
   #include &lt;iostream&gt;
   #include &lt;iomanip&gt;
   using namespace std;
#else
   #include &lt;iostream.h&gt;
   #include &lt;iomanip.h&gt;
#endif


<font color=#992233>/*----------------- beginning of improvization algorithms ---------------*/</font>


MidiMessage message;            <font color=#229933>// for extracting notes from the synthesizer</font>

CircularBuffer&lt;char&gt; notes;     <font color=#229933>// storage for notes being played</font>
CircularBuffer&lt;long&gt; times;     <font color=#229933>// storage for note times being played</font>
double occurances[12] = {0};    <font color=#229933>// number of notes occurances</font>
int fadeNote = 0;               <font color=#229933>// next to to go out of scope</font>
int fadeTime = 0;               <font color=#229933>// next time to go out of scope</font>
SigTimer metronome;             <font color=#229933>// for display period of key analysis</font>
double analysisDuration;        <font color=#229933>// duration in seconds of analysis window</font>
double tempo;                   <font color=#229933>// tempo of the keyanalysis</font>
int keyoctave = 7;              <font color=#229933>// the analysis key performance octave</font>
int displayKey2 = 0;            <font color=#229933>// display the second key possibility</font>

double majorKey[12];            <font color=#229933>// weights for analysis of major keys</font>
double minorKey[12];            <font color=#229933>// weights for analysis of minor keys</font>

Voice  firstVoice;              <font color=#229933>// for primary key voice</font>
Voice  secondVoice;             <font color=#229933>// for secondary key voice</font>
double offFraction = 0.75;      <font color=#229933>// for turning off next keyvoice.</font>

<font color=#229933>// function declarations:</font>
void        <a href=#analyzekey>analyzekey</a>(void);
void        <a href=#storeNote>storeNote</a>(char storeNote, long storeTime);
const char* <a href=#name>name</a>(int pitch_class);


<font color=#992233>/*--------------------- maintenance algorithms --------------------------*/</font>


<font color=#229933>//////////////////////////////</font>
<font color=#229933>//</font>
<font color=#229933>// description -- this function is called by the improv interface</font>
<font color=#229933>//     whenever a capital "D" is pressed on the computer keyboard.</font>
<font color=#229933>//     Put a description of the program and how to use it here.</font>
<font color=#229933>//</font>

void<a name=description> </a><font color=#000000><b>description</b></font>(void) {
   cout &lt;&lt; 
   "Keyan 1 -- analyze musical input for key\n"
   &lt;&lt; endl;

} 



<font color=#229933>//////////////////////////////</font>
<font color=#229933>//</font>
<font color=#229933>// initialization -- this function is called by the improv</font>
<font color=#229933>//     intervace once at the start of the program.  Put items</font>
<font color=#229933>//     here which need to be initialized at the beginning of</font>
<font color=#229933>//     the program.</font>
<font color=#229933>//</font>

void<a name=initialization> </a><font color=#000000><b>initialization</b></font>(void) { 
   analysisDuration = 7.0;          <font color=#229933>// duration in seconds of analysis window</font>
   tempo = 60.0;                     <font color=#229933>// tempo of the keyanalysis</font>
   metronome.setTempo(tempo);
   firstVoice.setChannel(0);
   secondVoice.setChannel(0);
   notes.setSize(10000);
   times.setSize(10000);

   <font color=#229933>// initialize majorKey and minorKey arrays</font>
   <font color=#229933>// The tonality coefficients listed here were determined by </font>
   <font color=#229933>// Krumhansl & Kessler (1982) and are defined for both </font>
   <font color=#229933>// the major and minor keys.</font>
   majorKey[0]  = 6.35;       minorKey[0]  = 6.33;
   majorKey[1]  = 2.23;       minorKey[1]  = 2.68;
   majorKey[2]  = 3.48;       minorKey[2]  = 3.52;
   majorKey[3]  = 2.33;       minorKey[3]  = 5.38;
   majorKey[4]  = 4.38;       minorKey[4]  = 2.60;
   majorKey[5]  = 4.09;       minorKey[5]  = 3.53;
   majorKey[6]  = 2.52;       minorKey[6]  = 2.54;
   majorKey[7]  = 5.19;       minorKey[7]  = 4.75;
   majorKey[8]  = 2.39;       minorKey[8]  = 3.98;
   majorKey[9]  = 3.66;       minorKey[9]  = 2.69;
   majorKey[10] = 2.29;       minorKey[10] = 3.34;
   majorKey[11] = 2.88;       minorKey[11] = 3.17;
}



<font color=#229933>//////////////////////////////</font>
<font color=#229933>//</font>
<font color=#229933>// finishup -- this function is called by the improv interface</font>
<font color=#229933>//     whenever the program is exited.  Put items here which</font>
<font color=#229933>//     need to be taken care of when the program is finished.</font>
<font color=#229933>//</font>

void<a name=finishup> </a><font color=#000000><b>finishup</b></font>(void) { 
   firstVoice.off();
   secondVoice.off();
}


<font color=#992233>/*-------------------- main loop algorithms -----------------------------*/</font>

<font color=#229933>//////////////////////////////</font>
<font color=#229933>//</font>
<font color=#229933>// mainloopalgorithms -- this function is called by the improv interface</font>
<font color=#229933>//   continuously while the program is running.  The global variable t_time</font>
<font color=#229933>//   which stores the current time is set just before this function is</font>
<font color=#229933>//   called and remains constant while in this functions.</font>
<font color=#229933>//</font>

void<a name=mainloopalgorithms> </a><font color=#000000><b>mainloopalgorithms</b></font>(void) { 
   while (synth.getNoteCount() &gt; 0) {
      message = synth.extractNote();
      if (message.p2() != 0) {
         <a href=#storeNote>storeNote</a>(message.p1(), t_time);
      }
   }

   if (metronome.expired()) {
      <a href=#analyzekey>analyzekey</a>();
      metronome.reset();
   }

   if (metronome.getPeriodCount() &gt; offFraction) {
      firstVoice.off();
      secondVoice.off();
   }

}
      

<font color=#992233>/*-------------------- triggered algorithms -----------------------------*/</font>


<font color=#229933>///////////////////////////////</font>
<font color=#229933>//</font>
<font color=#229933>// keyboardchar -- this function is called by the improv interface</font>
<font color=#229933>//     whenever a key is pressed on the computer keyboard.</font>
<font color=#229933>//     Put commands here which will be executed when a key is</font>
<font color=#229933>//     pressed on the computer keyboard.</font>
<font color=#229933>//</font>

void<a name=keyboardchar> </a><font color=#000000><b>keyboardchar</b></font>(int key) { 
   switch (key) {

      case '2':            <font color=#229933>// for playing the secondary key possibility</font>
         displayKey2 = !displayKey2;         
         if (displayKey2) {
            cout &lt;&lt; "Secondary key output turned ON" &lt;&lt; endl;
         } else {
            cout &lt;&lt; "Secondary key output turned OFF" &lt;&lt; endl;
         }
         break;

      case ',':            <font color=#229933>// slow down the tempo which will be</font>
         tempo /= 1.05;    <font color=#229933>// activated at the next note to be played</font>
         if (tempo &lt; 20) {
            tempo = 20;
         }
         metronome.setTempo(tempo);
         cout &lt;&lt; "Tempo is: " &lt;&lt; tempo &lt;&lt; endl;
         break;

      case '.':            <font color=#229933>// speed up the tempo which will be</font>
         tempo *= 1.05;    <font color=#229933>// activated at the next note to be played</font>
         if (tempo &lt; 240) {
            tempo = 240;
         }
         metronome.setTempo(tempo);
         cout &lt;&lt; "Tempo is: " &lt;&lt; tempo &lt;&lt; endl;
         break;

   }
}


<font color=#992233>/*------------------ begining of assistant functions --------------------*/</font>



<font color=#229933>//////////////////////////////</font>
<font color=#229933>//</font>
<font color=#229933>// analyzekey -- figure out the key possibilities and then play the</font>
<font color=#229933>//   best two choices.</font>
<font color=#229933>//</font>

void<a name=analyzekey> </a><font color=#000000><b>analyzekey</b></font>(void) { 
   int i;

   <font color=#229933>// first, adjust the number of notes being analyzed</font>
   while (times.getCount() &gt; 0 && fadeTime &lt; t_time - analysisDuration * 1000) {
      if (fadeTime != 0) {
         occurances[fadeNote%12]--;
      }
      fadeTime = times.extract();
      fadeNote = notes.extract();
   }
      
   <font color=#229933>// now analyze the keys</font>
   double total = 0;
   cout &lt;&lt; "Count:";
   for (i=0; i&lt;12; i++) {
      cout.width(3);
      cout &lt;&lt; occurances[i];
      total += occurances[i];
   }
   cout &lt;&lt; flush;
   
   <font color=#229933>// Determine the mean frequency for the pitch distribution</font>
   double mean       = total / 12.0;
   double major_mean = 3.4825;
   double minor_mean = 3.709167;

   if (total &lt;= 0) {
      cout &lt;&lt; endl;
      return;
   }
      
   double r_major[12];
   double r_minor[12];
   double maj_numerator;
   double min_numerator;
   double maj_denominator;
   double min_denominator;
   double maj_temp;
   double min_temp;
   double temp;
   int subscript;
   int j;
   for (i=0; i&lt;12; i++) {
      maj_numerator = min_numerator = 0;
      maj_denominator = min_denominator = 0;
      maj_temp = min_temp = temp = 0;

      <font color=#229933>// Examine all pitches for each key.</font>
      for (j=0; j&lt;12; j++) {
         subscript = (i+j) % 12;

         temp += (occurances[subscript] - mean) *
                 (occurances[subscript] - mean);

         <font color=#229933>// For the major keys.</font>
         maj_numerator += (majorKey[j]-major_mean) * 
                          (occurances[subscript]-mean);
         maj_temp += (majorKey[j] - major_mean) * 
                     (majorKey[j] - major_mean);
         
         <font color=#229933>// For the minor keys.</font>
         min_numerator += (minorKey[j]-minor_mean) * 
                          (occurances[subscript]-mean);
         min_temp += (minorKey[j] - minor_mean) *
                     (minorKey[j] - minor_mean);

      }

      maj_denominator = sqrt(maj_temp * temp);
      min_denominator = sqrt(min_temp * temp);
      
      if (maj_denominator == 0 || min_denominator == 0) {
         return;
      }

      r_major[i] = maj_numerator / maj_denominator;
      r_minor[i] = min_numerator / min_denominator;
   }

   <font color=#229933>// Now determine which correlation is the greatest.</font>
   <font color=#229933>// Start off with the assumption that C major is the best key.</font>
   double best_key = 0.0; 
   char* mode = "unknown";
   int pitch_class = 0;            <font color=#229933>// tonic note of best key</font>
   
   <font color=#229933>// Compare all the remaining key correlations.</font>
   for (i=0; i&lt;12; i++) {
      if (r_major[i] &gt; best_key) {
         best_key = r_major[i]; 
         mode = "major"; 
         pitch_class = i;
      }
      if (r_minor[i] &gt; best_key) {
         best_key = r_minor[i]; 
         mode = "minor"; 
         pitch_class = i;
      }
   }

   <font color=#229933>// A confidence measure can be determined by taking the difference</font>
   <font color=#229933>// between the correlation for the "best key" and subtracting the</font>
   <font color=#229933>// correlation for the "second best key".  The maximum confidence</font>
   <font color=#229933>// score is 100; the minimum is zero.</font>
   <font color=#229933>// First, having found the "best key", find the "second best key."</font>
   double second_best_key = 0;
   char* secondmode = "unknown";
   int sec_pitch_class = 0;
   for (i=0; i&lt;12; i++) {
      if (r_major[i] != best_key && r_major[i] &gt; second_best_key) {
         second_best_key = r_major[i];
         secondmode = "major";
         sec_pitch_class = i;
      }
      if (r_minor[i] != best_key && r_minor[i] &gt; second_best_key) {
         second_best_key = r_minor[i];
         secondmode = "minor";
         sec_pitch_class = i;
      }
   }

   <font color=#229933>// The value 3.0 below is a scaling factor.</font>
   double confidence = (best_key - second_best_key) * 100 * 3.0;
   if (confidence &gt; 100.0) confidence = 100.0;
   
   <font color=#229933>// Print the analysis results:</font>
   cout &lt;&lt; "   Key: " &lt;&lt; <a href=#name>name</a>(pitch_class) &lt;&lt; " " &lt;&lt; mode 
        &lt;&lt; " (";
   cout.width(3);
   cout &lt;&lt; (int)confidence &lt;&lt; "%), or " 
        &lt;&lt; <a href=#name>name</a>(sec_pitch_class) &lt;&lt; " " &lt;&lt; secondmode &lt;&lt; endl;

   <font color=#229933>// finally, play the determined keys according to confidence</font>
   int velocity = (int)(127.0 * confidence/100.0);
   
   firstVoice.play(pitch_class + 12 * keyoctave, velocity);
   if (displayKey2) {
      secondVoice.play(sec_pitch_class + 12 * keyoctave, 127 - velocity);
   }

}


<font color=#229933>//////////////////////////////</font>
<font color=#229933>//</font>
<font color=#229933>// name -- return the name of the input pitch class</font>
<font color=#229933>//</font>

const char*<a name=name> </a><font color=#000000><b>name</b></font>(int pitch_class) {
   pitch_class = pitch_class % 12;
   switch (pitch_class) {
      case 0:   return "C ";
      case 1:   return "C#";
      case 2:   return "D ";
      case 3:   return "Eb";
      case 4:   return "E ";
      case 5:   return "F ";
      case 6:   return "F#";
      case 7:   return "G ";
      case 8:   return "Af";
      case 9:   return "A ";
      case 10:  return "Bf";
      case 11:  return "B ";
   }
   
   return "unknown";
}
   




<font color=#229933>//////////////////////////////</font>
<font color=#229933>//</font>
<font color=#229933>// storeNote -- store the next note</font>
<font color=#229933>//</font>

void<a name=storeNote> </a><font color=#000000><b>storeNote</b></font>(char storeNote, long storeTime) {
   notes.insert(storeNote);
   times.insert(storeTime);
   occurances[storeNote%12]++;
}



<font color=#992233>/*------------------ end improvization algorithms -----------------------*/</font>



<font color=#229933>// md5sum: b0b4ae9d680c6bad68797d8fc2098132 keyan.cpp [20050403]</font>
</pre>
</body>
</html>
   
// md5sum: 819dc9944ce3cd34291de8c9ff010a93 rtkey.cpp [20090615]
